#include "xnflib.h"
#include <string.h>

int xnfTestWrite(const char*p){
	xmlDocPtr ref=0;
	int dims[3]={10,20,0};
	double*pd=0;
	int err=xnfCreate(p,&ref);
	if(err){
		printf("Error %d while creating the xnf file %s\n",err,p);
		return err;
	}
	int n=dims[0]*dims[1];
	pd=(double*)malloc(n*sizeof(double));
	for(int i=0;i<n;i++)pd[i]=i;
	xmlNodePtr dataset=xnfCreateDataset( ref, "essai2d",2, dims);
	if(dataset){
		xnfAddArray(dataset,n,pd,TypeDouble,0,0);
		for(int i=0;i<dims[0];i++)pd[i]=i/(dims[0]-1.0);
		err= xnfAddScale(dataset, 1,dims[0], pd,TypeDouble);
		if(!err)
			err=xnfAddScaleAsRange( dataset, 2, 0.0 , 0.3);
		//if(!err)
		err=xnfSave( ref);
	}else
		printf("Error while creating the dataset '%s'\n","essai2d");
	free(pd);
	xnfClose(ref);
	return err;
}

int xnfTestRead(const char*p){
	xmlDocPtr ref=0;
	int dim;
	int dims[3];
	int err=xnfOpen(p,&ref);
	if(err)return err;
	xmlNodePtr dataset=xnfGetDataset(ref, "essai2d");
	if(dataset){
		float*pd;
		err=xnfArrayGetContents(dataset, 1,&dim,dims,(void**)&pd,TypeFloat);
		if(err) printf("Error %d while reading contents of the array '%s'\n",err,"essai2d");
		int size=1;
		if(!err){
			for(int i=0;i<dim;i++)size*=dims[i];
			printf("contents: ");
			for(int i=0;i<size;i++) printf(" %g",pd[i]);
			printf("\n");free(pd);
		}
		err=xnfArrayGetScale(dataset, 1,&size, (void**)&pd,TypeFloat);
		if(!err){
			printf("scales1: ");
			for(int i=0;i<size;i++)printf(" %g",pd[i]);
			printf("\n");free(pd);
		}
		else
			printf("Error while reading scales 1 of the array '%s'\n","essai2d");

		err=xnfArrayGetScale(dataset, 2,&size, (void**)&pd,TypeFloat);
		if(!err){
			printf("scales2: ");
			for(int i=0;i<size;i++)printf(" %g",pd[i]);
			printf("\n");free(pd);
		}
		else
			printf("Error while reading scales 2 of the array '%s'\n","essai2d");
	}
	else
		printf("Error while reading the array '%s'\n","essai2d");
	xnfClose(ref);
	return err;
}


int xnfTestWriteRef(const char*p){
	xmlDocPtr ref=0;
	int err=xnfOpen(p,&ref);
	if(err)return err;
	int dim;
	int dims[3]={20,0,0};
	xmlNodePtr dataset=xnfCreateDataset( ref, "essai1din2d",1, dims);
	if(!dataset)
		printf("Error while adding the array '%s'\n","essai1din2d");
	else
		xnfAddByRef(dataset,"essai2d",8*10,TypeDouble);
	dataset=xnfGetDataset(ref, "essai1din2d");
	if(dataset){
		double*pd;
		err=xnfArrayGetContents(dataset,1, &dim,dims,(void**)&pd,TypeDouble);
		int size=1;
		if(!err){
			for(int i=0;i<dim;i++)size*=dims[i];
			printf("contents ref: ");
			for(int i=0;i<size;i++)printf(" %g",pd[i]);
			printf("\n");free(pd);
		}
		else
			printf("Error while reading contents of the array '%s'\n","essai1din2d");
	}
	else
		printf("Error while reading the array '%s'\n","essai1din2d");
	err=xnfSave( ref);
	xnfClose(ref);
	return err;
}

void xnftest(const char*p){
	printf("xnfTestWrite\n");
	xnfTestWrite(p);
	printf("\nxnfTestRead\n");
	xnfTestRead(p);
	printf("\nxnfTestWriteRef\n");
	xnfTestWriteRef(p);
}

int main (int argc, char * const argv[]) {
    //putenv ("XML_CATALOG_FILES=file:///etc/xml/catalog  file:///Library/DTDs/catalog");//paths to the possible catalog files
	char *curpath;
	asprintf(&curpath,__FILE__);
	int len=strlen(curpath)-1;
	while(curpath[len]!='/'){curpath[len]=0;len--;}
	strcat(curpath,"test.xnf");
	xnftest(curpath);//create a file "test.xnf" in the project directory
	free(curpath);
	return 0;
}
