/*
 *  xnflib.cpp
 *
 *  Created by Fran√ßois Delyon on 03.03.06.
 *  Copyright 2006 Satimage. All rights reserved.
 *
 */

#include "xnflib.h"
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <libkern/OSByteOrder.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

int xnfTypeSize[10]={8,4,4,2,1,4,2,1,16,8};

int xnfGetDataText(char* s,int size,void*p,NumberType floatType);
int xnfReadBlock(char * xnfpath,int offset,NumberType infloatType,int inendian,int size,void* pp,NumberType outfloatType,int outendian);

int xnfWriteBlock(char* tempstr,void*pp, int size,NumberType isfloat,fpos_t* pos,NumberType outType,int outEndian);
int xnfGetDataNode(xmlNodePtr node,int index,xmlNodePtr* data);
int xnfGetAxisNode(xmlNodePtr node,int rank,xmlNodePtr* axis);
void TranslateData(int size,const void*inp,NumberType infloatType,int inendian,void*outp,NumberType outfloatType,int outendian);
typedef int (*xnfReadBlockFunc)(char * xnfpath,int offset,NumberType infloatType,int inendian,int size,void* pp,NumberType outfloatType,int outendian);
xnfReadBlockFunc gxnfReadBlockFunc=xnfReadBlock;
int HostIsBigEndian(){
	int i=1;
	return ((char*)&i)[3]==1;
}
int currentByteOrder=HostIsBigEndian();

static char* PathFromURL(char* p){
	char * filename=p;
    if (!xmlStrncasecmp(BAD_CAST p, BAD_CAST "file://localhost/", 17))
		filename = &p[16];
    else if (!xmlStrncasecmp(BAD_CAST p, BAD_CAST "file:///", 8)) 
		filename = &p[7];
	return xmlURIUnescapeString(filename, 0, 0);
}

xmlChar * PathFromHref(xmlNodePtr cur){
	xmlChar* href=xmlGetProp(cur, (const xmlChar *)"href");
	if(!href)return 0;
	xmlChar *base = xmlNodeGetBase(cur->doc, cur);
	xmlChar *ret=xmlBuildURI(href,base);
	xmlFree(href);
	if(base)xmlFree(base);
	xmlChar * filename=ret;
    if (!xmlStrncasecmp(BAD_CAST ret, BAD_CAST "file://localhost/", 17))
		filename = &ret[16];
    else if (!xmlStrncasecmp(BAD_CAST ret, BAD_CAST "file:///", 8)) 
		filename = &ret[7];
	filename=(xmlChar *)PathFromURL((char*)filename);
	xmlFree(ret);
	return filename;
}

static void xnfSetBase(xmlDocPtr doc){
	xmlNodePtr cur=xmlDocGetRootElement(doc);
	xmlChar *base = xmlGetNsProp(cur, BAD_CAST "base", XML_XML_NAMESPACE);
	if(!base){
		xmlNodeSetBase( cur, (const xmlChar *)"Contents/");
	}
	else
		xmlFree(base);
}

int xnfOpen(const char* filename,xmlDocPtr* ref){
	int err=0;
	*ref=0;
	if(!filename)return 0;
	xmlKeepBlanksDefault(0);
	xmlChar * xnfpath= xmlStrdup((const xmlChar *)filename);
	if(xnfpath[xmlStrlen(xnfpath)-1]!='/')
		xnfpath=xmlStrcat(xnfpath,(xmlChar *)"/");
	xmlChar * index= xmlStrcat(xnfpath,(xmlChar *)"index.xml");
	xmlDoValidityCheckingDefaultValue=1;
	xmlDocPtr doc=xmlReadFile((char*)index,0, XML_PARSE_RECOVER|XML_PARSE_NOBLANKS);//XML_PARSE_DTDLOAD|XML_PARSE_DTDVALID
	xmlFree(index);
	if(!doc){
		xmlErrorPtr xmlerr=xmlGetLastError();
		if(xmlerr && xmlerr->code != XML_ERR_OK)
			err=xmlerr->code;
		else err= -1;
		return err;
	}
	*ref=doc;
	xnfSetBase( doc);
	return 0;
}

void xnfClose(xmlDocPtr doc){
	if(!doc)return;
	xmlFreeDoc(doc);
}

int xnfSave(xmlDocPtr doc){
	if(!doc)return -1;
	char *file=(char*)doc->URL;
	if(!file)return -1;
	int ret= xmlSaveFormatFile(file, doc,1) ;
	char* pp=file+strlen(file)-1;
	for(int i=strlen(file);i>0;pp--,i--)
		if(*pp=='/')break;
	if(*pp=='/'){
		*pp=0;
		char * uf=PathFromURL(file);
		utimes(uf,0);
		*pp='/';free(uf);
	}
	return (ret<0)?ret:0;
}

xmlNodePtr xnfGetDataset(xmlDocPtr doc, const char* ID){
	if(!doc)return 0;
	xmlAttrPtr attr=xmlGetID(doc, (const xmlChar *)ID);
	if(!attr)return 0;
	return attr->parent;
}

static int xmlGetRealProp(xmlNodePtr cur, const char*prop,double*x){
	char *s=(char*) xmlGetProp(cur, (const xmlChar *)prop);
	if(!s)return -3;
	char* s1;
	*x=strtod(s,&s1);
	xmlFree(s);
	if(s==s1)return -1;
	return 0;
}

int xmlGetIntProp(xmlNodePtr cur, const char*prop,int*x){
	char *s=(char*) xmlGetProp(cur, (const xmlChar *)prop);
	if(!s)return -3;
	char* s1;
	*x=strtol(s,&s1,10);
	xmlFree(s);
	if(s==s1)return -1;
	return 0;
}

int xnfGetDimensions(xmlNodePtr node,int* dim,int* dims){
	if(!node || node->type!=XML_ELEMENT_NODE)return DimNotFound;
	if(xmlStrEqual(node->name, (const xmlChar *)"data"))
		node=node->parent;
	int err=xmlGetIntProp(node, "dimension",dim);
	if(err==-3)return DimNotFound;
	if(err || *dim<1 || *dim>MaxDim)return BadDim;
	xmlNodePtr cur = node->children;
	int size=-1;
	int nsizes=0;
    while (cur != NULL ) {
		if(xmlStrEqual(cur->name, (const xmlChar *)"axis")){
			err=xmlGetIntProp(cur, "size",&size);
			if(err==-3)return scale_sizeMissing;
			if(err || size<1 )return bad_size;
			if(nsizes<*dim)dims[nsizes]=size;
			nsizes++;
			if(nsizes==*dim)break;
		}
        cur = cur->next;
    }
	if(nsizes==*dim)return 0;
	return invalidNumberOfAxis;
}

static const char *typestrlist[]={"real64","real32","sint32","sint16","sint8","uint32","uint16","uint8","complex64","complex32"};

NumberType TypeFromStr(char*typestr){
	for(int i=0;i<10;i++)
		if(xmlStrEqual((const xmlChar *)typestr, (const xmlChar *)typestrlist[i]))
			return (NumberType)i;
	return (NumberType)-1;
}

int xnfDataGetInfo(xmlNodePtr cur,NumberType*type,int* byte_order,__int64_t *offset){
	*offset=0;*byte_order=true;*type=(NumberType)-1;
	xmlChar* typestr=xmlGetProp(cur, (const xmlChar *)"type");
	if(!typestr)return -2;
	*type=TypeFromStr((char*)typestr);
	xmlFree(typestr);
	if(*type<0) return unknownElementType;
	char* offsetstr=(char*)xmlGetProp(cur, (const xmlChar *)"offset");
	if(offsetstr){
		*offset=atoll(offsetstr);
		xmlFree(offsetstr);
	}
	xmlChar* endian=xmlGetProp(cur, (const xmlChar *)"byte_order");
	if(!endian)endian=xmlGetProp(cur, (const xmlChar *)"endian");
	if(!endian)return endianessMissing;
	if(xmlStrEqual(endian, (const xmlChar *)"little"))*byte_order=0;
	xmlFree(endian);
	return 0;
}

int xnfDataGetContents(xmlNodePtr cur,int size,void**p,NumberType floatType,int doAlloc){
	__int64_t offset=0;
	int err=0;
	NumberType infloatType;int byte_order;
	xmlChar * xnfpath= PathFromHref(cur);
	if(!xnfpath){
		char* s=(char*) xmlNodeGetContent( cur);
		if(s){
			if(doAlloc)*p=malloc(size*xnfTypeSize[floatType]);
			if(!*p)
				err=memFull;
			else{
				err=xnfGetDataText(s, size,*p, floatType);
				if(err && doAlloc){free(*p);*p=0;}
			}
			xmlFree(s);
			return err;
		}
		return cantGetPath;
	}
	err=xnfDataGetInfo( cur, &infloatType,&byte_order ,&offset);
	if(!err){
		if(doAlloc)*p=malloc(size*xnfTypeSize[floatType]);
		if(!*p)
			err=memFull;
		else{
			err= gxnfReadBlockFunc(( char * )xnfpath, offset,infloatType,byte_order,size,*p,floatType,currentByteOrder);
			if(err && doAlloc){free(*p);*p=0;}
		}
	}
	xmlFree(xnfpath);
	return err;
}

static int xnfArrayCanRemoveFile(xmlXPathContextPtr xpathCtx,xmlNodePtr node){
	int ret=0;
	xmlChar* href=xmlGetProp(node, (const xmlChar *)"href");
	if(!href)return -1;
	xmlXPathCompExprPtr exp=0;
	xmlXPathObjectPtr value=xmlXPathNewString(href);
	int err=xmlXPathRegisterVariable(xpathCtx, (const xmlChar *)"href", value);
	if(!err){
		exp=xmlXPathCtxtCompile(xpathCtx,(const xmlChar *)"//data[@href=$href]");
		if(exp){
			xmlXPathObjectPtr xpathObj = xmlXPathCompiledEval(exp, xpathCtx);
			xmlXPathFreeCompExpr(exp);
			if(xpathObj){
				if(xpathObj->nodesetval->nodeNr==1)ret=1;
				xmlXPathFreeObject(xpathObj);
			}
		}
	}
	return ret;
}

static int xnfDataRemoveWithXPath(xmlNodePtr cur,xmlXPathContextPtr xpathCtx){
	if(xnfArrayCanRemoveFile(xpathCtx, cur)){
		xmlChar * path=PathFromHref(cur);
		if(path){
			unlink((char*)path);
			free(path);
		}
	}
	xmlUnlinkNode(cur);
	xmlFreeNode(cur);
	return 0;
}

static int xnfDataRemove(xmlNodePtr node){
	xmlXPathContextPtr xpathCtx=xmlXPathNewContext(node->doc);
	if(!xpathCtx)return -1;
	xnfDataRemoveWithXPath( node, xpathCtx);
	xmlXPathFreeContext(xpathCtx);
	return 0;
}
	
int xnfDatasetRemove(xmlNodePtr node,int del ){
	xmlXPathContextPtr xpathCtx=xmlXPathNewContext(node->doc);
	if(xpathCtx){
		for(;;){
			xmlNodePtr cur=0;
			int err=xnfGetDataNode( node,1, &cur);
			if(err)break;
			xnfDataRemoveWithXPath( cur, xpathCtx);
		}
		xmlXPathFreeContext(xpathCtx);
	}
	if(del){
		xmlUnlinkNode(node);
		xmlFreeNode(node);
	}
	return 0;
}

int xnfRemove(xmlNodePtr node){
	if(!node)return -1;
	if(node->type!=XML_ELEMENT_NODE)return -1;
	if(xmlStrEqual(node->name, (const xmlChar *)"data"))
		return xnfDataRemove( node);
	else if(xmlStrEqual(node->name, (const xmlChar *)"dataset"))
		return xnfDatasetRemove( node,1);
	xmlUnlinkNode(node);
	xmlFreeNode(node);
	return 0;
}

int xnfArrayGetContentsAlloc(xmlNodePtr node,int index,int* dim,int* dims,void**p,NumberType floatType,int doAlloc){
	if(index<1)return IllegalDataIndex;
	int err=xnfGetDimensions(node,dim, dims);
	if(err)return err;
	xmlNodePtr data=0;
	err=xnfGetDataNode( node,index, &data);
	if(err)return err;
	int size=1;
	for(int i=0;i<*dim;i++)size*=dims[i];
	err= xnfDataGetContents(data,  size,p,floatType,doAlloc);
	return err;
}

int xnfArrayGetContents(xmlNodePtr node,int index,int* dim,int* dims,void**p,NumberType floatType){
	return xnfArrayGetContentsAlloc( node,index,  dim, dims,(void**)p, floatType,1);
}

int xnfArrayGetContentsMem(xmlNodePtr node,int index,void*p,NumberType floatType){
	void* q=p;
	int dim, dims[MaxDim];
	return xnfArrayGetContentsAlloc( node,index,&dim, dims,(void**)&q,floatType,0);
}

int xnfArrayGetScaleRange(xmlNodePtr cur,int size,void**p,int floatType,int doAlloc){
	if(floatType<0 ||floatType>1 )return invalidType;
	double start,step;
	int err=xmlGetRealProp(cur,"start",&start);
	if(err)return err;
	err=xmlGetRealProp(cur,"step",&step);
	if(err)return err;
	if(floatType==TypeFloat){
		if(doAlloc)
			*p=malloc(size*sizeof(float));
		float* pp=(float*)*p;
		for(int i=0;i<size;i++,pp++)
			*pp=start+i*step;
	}else{
		if(doAlloc)
			*p=malloc(size*sizeof(double));
		double* pp=(double*)*p;
		for(int i=0;i<size;i++,pp++)
			*pp=start+i*step;
	}
	return 0;
}

int xnfArrayGetScaleAlloc(xmlNodePtr cur,int rank,int& size,void**p,NumberType floatType,int doAlloc){
	xmlNodePtr axis=0;
	int err=xnfGetAxisNode( cur, rank,&axis);
	if(err)return err;
	err=xmlGetIntProp(axis, "size",&size);
	if(err==-3)return scale_sizeMissing;
	if(err || size<1)return bad_size;
	if(err)return err;
	char *s=(char*) xmlGetProp(axis, (const xmlChar *)"idref");
	if(s){
		xmlNodePtr node=xnfGetDataset( cur->doc, s);
		xmlFree(s);
		if(!node)
			return invalid_axisIdRef;
		int dim;int dims[MaxDim];
		err=xnfArrayGetContentsAlloc( node,1,&dim,dims,p,floatType,doAlloc);
		if(err)return err;
		if(dim>1 || dims[0]!=size){
			if(doAlloc)free(*p);return invalid_axisIdRef;
		}
		return 0;	
	}
	err= xnfArrayGetScaleRange(axis,size,p,floatType,doAlloc);
	if(!err)return 0;
	xmlNodePtr data=0;
	err= xnfGetDataNode(axis,1,&data);
	if(err)return err;
	err=xnfDataGetContents(data,size,p, floatType,doAlloc);
	return err;
}

int xnfArrayGetScale(xmlNodePtr cur,int rank,int* size,void**p,NumberType floatType){
	return xnfArrayGetScaleAlloc( cur, rank,*size,p, floatType,1);
}

int xnfArrayGetScaleMem(xmlNodePtr node,int rank,void*p,NumberType floatType){
	int n;
	void*q=p;
	return xnfArrayGetScaleAlloc( node, rank, n,&q,floatType,0);
}

int xnfCreate(const char* xnfpath,xmlDocPtr* ref){
	FILE*f=0;
	int len=strlen(xnfpath);
	int	err=mkdir(xnfpath, 0777);
	if(err)return err;
	char* s=(char*)malloc(len+20);
	memcpy(s,xnfpath,len+1);
	strcat(s,"/Contents");
	err=mkdir(s, 0777);
	if(err)return errno;
	memcpy(s,xnfpath,len+1);
	strcat(s,"/index.xml");
	f=fopen(s,"w");
	fputs("<?xml version='1.0' encoding='UTF-8'?>\n<!DOCTYPE tableofcontents SYSTEM \"http://www.satimage-software.com/dtds/XNFv2.dtd\">\n<tableofcontents xml:base='Contents/'/>",f);
	fclose(f);
	return xnfOpen(xnfpath, ref);
}

xmlNodePtr GetLastNode(xmlNodePtr node){
	xmlNodePtr cur = node->last;
    while (cur != NULL ) {
		if(xmlStrEqual(cur->name, (const xmlChar *)"axis"))
			return cur;
		if(xmlStrEqual(cur->name, (const xmlChar *)"data"))
			return cur;
        cur = cur->prev;
    }
	return 0;
}

xmlNodePtr xnfAddData(xmlNodePtr cur,const char* href,int offset,NumberType floatType,int byte_order){
	if(xmlStrEqual(cur->name, (const xmlChar *)"data"))
		cur=cur->parent;
	xmlNodePtr prev=GetLastNode(cur);
	xmlChar *ID=0;
	if(!href){
		ID=xmlGetProp(cur,(const xmlChar *)"id");
		if(!ID && cur->parent)
			ID=xmlGetNsProp(cur->parent,(const xmlChar *)"id",XML_XML_NAMESPACE);
		if(!ID)return 0;
	}
	xmlNodePtr data= xmlNewNode(0, (const xmlChar *)"data");
	if(prev)
		xmlAddNextSibling(prev, data);
	else
		xmlAddChild(cur,  data);
	if(href)
		xmlSetProp(data, (const xmlChar *)"href",(const xmlChar *)href);
	else{
		xmlChar * escapedId=xmlURIEscapeStr((const xmlChar *)ID,0);
		xmlFree(ID);
		xmlSetProp(data, (const xmlChar *)"href",(const xmlChar *)escapedId);
		xmlFree(escapedId);
	}
	char tempstr[50];
	sprintf(tempstr,"%d",offset);
	xmlSetProp(data, (const xmlChar *)"offset",(const xmlChar *)tempstr);
	xmlSetProp(data, (const xmlChar *)"type",(const xmlChar *)typestrlist[floatType]);
	if(byte_order)
		xmlSetProp(data, (const xmlChar *)"byte_order",(const xmlChar *)"big");
	else
		xmlSetProp(data, (const xmlChar *)"byte_order",(const xmlChar *)"little");
	return data;
}

xmlNodePtr xnfCreateDataset(xmlDocPtr ref, const char* ID,int dim,int dims[MaxDim]){
	if(dim<1 ||dim>MaxDim)return 0;
	xmlNodePtr prev=xnfGetDataset(ref,ID);
	if(prev)return 0;
	char temp[256];
	xmlNodePtr cur= xmlNewNode(0, (const xmlChar *)"dataset");
	if(ref->type==XML_ELEMENT_NODE)
		xmlAddChild((xmlNodePtr)ref,  cur);
	else
		xmlAddChild(xmlDocGetRootElement(ref),  cur);
	xmlNsPtr	ns = xmlSearchNsByHref(cur->doc, cur, XML_XML_NAMESPACE);
	xmlAttrPtr attr=xmlSetNsProp(cur,ns, (const xmlChar *)"id",(const xmlChar *)ID);
	if(attr)xmlAddID(NULL, cur->doc, (const xmlChar *)ID, attr);
	//xmlSetProp(cur, (const xmlChar *)"xml:id",(const xmlChar *)ID);
	sprintf(temp,"%d",dim);
	xmlSetProp(cur, (const xmlChar *)"dimension",(const xmlChar *)temp);
	for(int i=0;i<dim;i++){
		xmlNodePtr axis= xmlNewNode(0, (const xmlChar *)"axis");
		xmlAddChild(cur,  axis);
		sprintf(temp,"%d",dims[i]);
		xmlSetProp(axis, (const xmlChar *)"size",(const xmlChar *)temp);
	}
	return cur;
}

int xnfAddArrayTransLate(xmlNodePtr cur,int size,void*pp,NumberType floatType,const char*href,NumberType outtype,int outEndian,xmlNodePtr*dataout){
	if(outtype==-1)outtype=floatType;
	if(outtype<=TypeUChar && floatType>TypeUChar)
		return incompatibleType;
	xmlNodePtr data=xnfAddData( cur,href,0, outtype,outEndian);
	if(dataout)*dataout=0;
	if(!data)return -1;
	fpos_t pos=0;
	int err=0;
	if(pp){
		xmlChar* tempstr=PathFromHref(data);
		if(tempstr){
			err=xnfWriteBlock((char*) tempstr,pp,  size,floatType,&pos,outtype, outEndian);
			xmlFree(tempstr);
		}
	}
	if(err){
		xmlUnlinkNode(data);xmlFreeNode(data);return err;
	}
	if(pos>0){
		char tempstr[50];
		sprintf(tempstr,"%ld",(long int)pos);
		xmlSetProp(data, (const xmlChar *)"offset",(const xmlChar *)tempstr);
	}
	if(dataout)*dataout=data;
	return err;
}

int xnfAddArray(xmlNodePtr cur,int size,void*pp,NumberType floatType,const char*href,xmlNodePtr*data){
	return xnfAddArrayTransLate( cur, size,pp, floatType,href,floatType,currentByteOrder,data);
}

xmlNodePtr xnfGetCleanAxisNode(xmlNodePtr cur,int rank){
	xmlNodePtr axis=0;
	xnfGetAxisNode( cur, rank,&axis);
	if(!axis)return 0;
	xmlUnsetProp(axis, (const xmlChar *)"step");
	xmlUnsetProp(axis, (const xmlChar *)"start"); 
	xmlUnsetProp(axis, (const xmlChar *)"idref");
	xmlFreeNodeList(axis->children);
	axis->last = axis->children = NULL;
	return axis;
}

void p2p(int size,const void*p, int intype,void*p2,int outtype);

int xnfAddCDATAScale(xmlNodePtr cur,int rank,int n,void*p,int floatType){
	xmlNodePtr axis= xnfGetCleanAxisNode(cur,rank);
	if(!axis)return -1;
	xmlNodePtr data= xmlNewNode(0, (const xmlChar *)"data");
	xmlAddChild(axis,data);
	int size=0;
	int allocated=n*20;
	char*s=(char*)malloc(allocated);
	*s=0;
	char ss[50];
	int *px=(int*)malloc(sizeof(int)*n);
	if(floatType>1) p2p(n,p, floatType,(void*)px,2);
	for(int i=0;i<n;i++){
		double x;
		if(floatType==1){
			x=((float*)p)[i];
			sprintf(ss,"%g ",x);
		}else if(floatType==0){
			x=((double*)p)[i];
			sprintf(ss,"%g ",x);
		}else
			sprintf(ss,"%d ",px[i]);
		int len=strlen(ss);
		if(size+len>=allocated){
			allocated+=100;
			s=(char*)realloc(s,allocated);
		}
		strcat(s,ss);size+=len;
	}
	xmlNodeSetContent(data,(const xmlChar *)s);
	free(s);
	return 0;
}

int xnfAddScaleTranslate(xmlNodePtr cur,int rank,int n,void*p,NumberType floatType,NumberType outType,int outendian){
	xmlNodePtr axis= xnfGetCleanAxisNode(cur,rank);
	if(!axis)return scale_sizeMissing;
	xmlNodePtr data= 0;
	return xnfAddArrayTransLate(axis,n,p, floatType,0,outType,outendian,&data);
}

int xnfAddScale(xmlNodePtr cur,int rank,int n,void*p,NumberType floatType){
	return xnfAddScaleTranslate( cur, rank, n,p, floatType,floatType,currentByteOrder);
}

int xnfAddScaleAsRange(xmlNodePtr cur,int rank,double start , double step){
	char tempstr[50];
	xmlNodePtr axis= xnfGetCleanAxisNode(cur,rank);
	if(!axis)return -1;
	sprintf(tempstr,"%g",start);
	xmlSetProp(axis, (const xmlChar *)"start",(const xmlChar *)tempstr);
	sprintf(tempstr,"%g",step);
	xmlSetProp(axis, (const xmlChar *)"step",(const xmlChar *)tempstr);
	return 0;
}

int xnfAddScaleAsRef(xmlNodePtr cur,int rank,const char* ID){
	xmlNodePtr axis= xnfGetCleanAxisNode(cur,rank);
	if(!axis)return -1;
	if(ID)xmlSetProp(axis, (const xmlChar *)"idref",(const xmlChar *)ID);
	return 0;
}

xmlNodePtr xnfAddByRef(xmlNodePtr cur,const char* filename,int offset,NumberType floatType){
	if(!cur)return 0;
	return xnfAddData(cur, filename,offset,  floatType,currentByteOrder);
}

void SwapArray(void*pp,int size,int elemsize){
	if(elemsize == sizeof(uint32_t)){
		uint32_t* p=(uint32_t*)pp;
		for(int i=0;i<size;p++,i++)*p=OSSwapInt32(*p);
	}else if(elemsize == sizeof(uint64_t)){
		uint64_t* p=(uint64_t*)pp;
		for(int i=0;i<size;p++,i++)*p=OSSwapInt64(*p);
	}
	else if(elemsize == sizeof(uint16_t)){
		uint16_t* p=(uint16_t*)pp;
		for(int i=0;i<size;p++,i++)*p=OSSwapInt16(*p);
	}
}

#define SWITCH		switch(intype){\
	case TypeDouble:{\
		double *inp=(double *)p;\
		for(int i=0;i<size;i++,inp++,outp++)*outp=*inp;\
		}break;\
	case TypeFloat:{\
		float *inp=(float *)p;\
		for(int i=0;i<size;i++,inp++,outp++)*outp=*inp;\
		}break;\
	case TypeShort:{\
		short *inp=(short *)p;\
		for(int i=0;i<size;i++,inp++,outp++)*outp=*inp;\
		}break;\
	case TypeUShort:{\
		unsigned short *inp=(unsigned short *)p;\
		for(int i=0;i<size;i++,inp++,outp++)*outp=*inp;\
		}break;\
	case TypeInteger:{\
		int *inp=(int *)p;\
		for(int i=0;i<size;i++,inp++,outp++)*outp=*inp;\
		}break;\
	case TypeUInteger:{\
		unsigned int *inp=(unsigned int *)p;\
		for(int i=0;i<size;i++,inp++,outp++)*outp=*inp;\
		}break;\
	case TypeChar:{\
		char *inp=(char *)p;\
		for(int i=0;i<size;i++,inp++,outp++)*outp=*inp;\
		}break;\
	case TypeUChar:{\
		unsigned char *inp=(unsigned char *)p;\
		for(int i=0;i<size;i++,inp++,outp++)*outp=*inp;\
		}break;\
	}\
	
void p2p(int size,const void*p, int intype,void*p2,int outtype){
	if(intype==outtype){
		memcpy(p2,p,size*xnfTypeSize[outtype]);
		return;
	}
	if(intype>=TypeDoubleComplex){
		intype-=TypeDoubleComplex;
		outtype-=TypeDoubleComplex;
		size*=2;
	}
	switch(outtype){
	case TypeDouble:{
		double *pp=(double*)p2;
		double *outp=pp;
		SWITCH;
		}break;
	case TypeFloat:{
		float *pp=(float*)p2;
		float *outp=pp;
		SWITCH;
		}break;
	case TypeShort:{
		short *pp=(short*)p2;
		short *outp=pp;
		SWITCH;
		}break;
	case TypeUShort:{
		unsigned short *pp=(unsigned short*)p2;
		unsigned short *outp=pp;
		SWITCH;
		}break;
	case TypeInteger:{
		int *pp=(int*)p2;
		int *outp=pp;
		SWITCH;
		}break;
	case TypeUInteger:{
		unsigned int *pp=(unsigned int*)p2;
		unsigned int *outp=pp;
		SWITCH;
		}break;
	case TypeChar:{
		char *pp=(char*)p2;
		char *outp=pp;
		SWITCH;
		}break;
	case TypeUChar:{
		unsigned char *pp=(unsigned char*)p2;
		unsigned char *outp=pp;
		SWITCH;
		}break;
	}
}

void TranslateData(int size,const void*inp,NumberType infloatType,int inendian,void*outp,NumberType outfloatType,int outendian){
	void*temp=0;
	if(inendian!=currentByteOrder){
		temp=malloc(xnfTypeSize[infloatType]*size);
		memcpy(temp,inp,xnfTypeSize[infloatType]*size);
		if(infloatType>=TypeDoubleComplex)
			SwapArray(temp,size*2,xnfTypeSize[infloatType]/2);
		else
			SwapArray(temp,size,xnfTypeSize[infloatType]);
		inp=temp;
	}
	p2p(size,inp,infloatType,outp,outfloatType);
	if(outendian!=currentByteOrder){
		if(outfloatType>=TypeDoubleComplex)
			SwapArray(outp,size*2,xnfTypeSize[outfloatType]/2);
		else
		SwapArray(outp,size,xnfTypeSize[outfloatType]);
	}
	if(temp)free(temp);
}

int xnfReadBlockf(FILE	*f,int offset,NumberType infloatType,int inendian,int size,void* pp,NumberType outfloatType,int outendian){
	if(outfloatType<TypeDoubleComplex && infloatType>=TypeDoubleComplex)
		return incompatibleType;
	if(outfloatType>=TypeDoubleComplex && infloatType<TypeDoubleComplex)
		return incompatibleType;
	int blocksize=4096;
	if(!size)return 0;
	int err=fseek ( f ,  offset , 0 );
	if(err)
		return cantOffset;
	if(blocksize>size)
		blocksize=size;
	void* buffer = malloc(blocksize*xnfTypeSize[infloatType]);
	while(size){
		int n,nn=(size>blocksize)?blocksize:size;
		n=fread ( buffer, xnfTypeSize[infloatType], nn, f);
		if(n!=nn)err=eoferror;
		if(err)break;
		size-=n;
		TranslateData(n,buffer, infloatType, inendian,pp, outfloatType, outendian);
		pp=(char*)pp+n*xnfTypeSize[outfloatType];
	}
	free(buffer);
	return err;
}

int xnfReadBlock(char * xnfpath,int offset,NumberType infloatType,int inendian,int size,void* pp,NumberType outfloatType,int outendian){
	FILE	*f=fopen(( char * )xnfpath, "r");
	if(!f)return cantOpen;
	int err=xnfReadBlockf(f, offset, infloatType, inendian, size, pp, outfloatType, outendian);
	fclose(f);
	return err;
}

int xnfWriteBlock(char* tempstr,void*pp, int size,NumberType infloatType,fpos_t* pos,NumberType outType,int outEndian){
	if(outType<=TypeUChar && infloatType>TypeUChar)
		return incompatibleType;
	if(outType>TypeUChar && infloatType<=TypeUChar)
		return incompatibleType;
	int blocksize=4096;
	FILE*f=0;
	int err=0;
	if(!pp || !size)return 0;
	f=fopen((char*)tempstr,"a");
	if(!f)return cantOpen;
	fgetpos(f, pos);
	if(infloatType==outType && outEndian==currentByteOrder){
		size_t n=fwrite (pp , xnfTypeSize[infloatType]  , size , f);
		fclose(f);
		if(n!=size_t(size))return cantWrite;
		return 0;
	}
	if(blocksize>size)
		blocksize=size;
	void* buffer = malloc(blocksize*xnfTypeSize[outType]);
	while(size){
		size_t nn=(size>blocksize)?blocksize:size;
		TranslateData(nn,pp, infloatType, currentByteOrder,buffer, outType, outEndian);
		pp=(char*)pp+nn*xnfTypeSize[infloatType];
		size_t n=fwrite (buffer , xnfTypeSize[outType]  , nn , f);
		if(n!=nn)err=eoferror;
		if(err)break;
		size-=n;
	}
	free(buffer);
	fclose(f);
	return err;
}

int xnfGetDataText(char* s,int size,void*p,NumberType floatType){
	if(!s)return -1;
	double*pd=(double*)malloc(size*sizeof(double));
	int i;
	for( i=0;i<size;i++){
		char* s1;
		while(*s==' ' || *s=='	')s++;
		if(*s==',')s++;
		if(!*s){free(pd);return scale_badstrdata;}
		pd[i]=strtod(s,&s1);
		if(s==s1){free(pd);return scale_badstrdata;}
		s=s1;
	}
	if(i!=size){free(pd);return scale_badstrdata;}
	p2p( size,pd, TypeDouble,p,floatType);
	free(pd);
	return 0;
}

int xnfGetDataNode(xmlNodePtr node,int index,xmlNodePtr* data){
	*data=0;
	if(xmlStrEqual(node->name, (const xmlChar *)"data")){
		*data=node;
		return 0;
	}
	if(index<1)return IllegalDataIndex;
	xmlNodePtr cur = node->children;
	int done=0;
    while (cur != NULL ) {
		if(xmlStrEqual(cur->name, (const xmlChar *)"data")){
			index--;done=1;
			if(index==0){
				*data=cur;
				return 0;
			}
		}
        cur = cur->next;
    }
	return (done)?IllegalDataIndex:dataMissing;
}

int xnfGetAxisNode(xmlNodePtr node,int index,xmlNodePtr* data){
	*data=0;
	xmlNodePtr cur = node->children;
	int done=0;
    while (cur != NULL ) {
		if(xmlStrEqual(cur->name, (const xmlChar *)"axis")){
			index--;done=1;
			if(index==0){
				*data=cur;
				return 0;
			}
		}
        cur = cur->next;
    }
	return (done)?IllegalAxisIndex:axisMissing;
}

int xnfReadRangeContentsf(FILE	*f,char *pp,NumberType infloatType,int inendian, NumberType outfloatType,int outendian,int dim,int*dims,int * lengths,fpos_t offset){
	if(dim==1)return xnfReadBlockf(f, offset, infloatType, inendian,lengths[0],pp, outfloatType, outendian);
	if(outfloatType<TypeDoubleComplex && infloatType>=TypeDoubleComplex)
		return incompatibleType;
	if(outfloatType>=TypeDoubleComplex && infloatType<TypeDoubleComplex)
		return incompatibleType;
	int blocksize=4096;
	int index[MaxDim];
	memset(index,0,MaxDim*sizeof(int));
	int err=fseek ( f ,  offset , SEEK_SET );
	if(err)
		return cantOffset;
	char* buffer = (char*)malloc(blocksize*xnfTypeSize[infloatType]);
	int nInBlock=0;
	int n=lengths[0];
	int cumulDims[MaxDim];
	int c=1;
	for(int i=0;i<dim;i++){cumulDims[i]=c;c*=dims[i];}
	for(;;){
		int nn=n;
		if(nn+nInBlock>blocksize)
			nn=blocksize-nInBlock;
		int dum=fread ( buffer+nInBlock*xnfTypeSize[infloatType], xnfTypeSize[infloatType], nn, f);
		if(dum!=nn)err=eoferror;
		if(err)break;
		n-=nn;
		nInBlock+=nn;
		if(nInBlock==blocksize){
			TranslateData(nInBlock,buffer, infloatType, inendian,pp, outfloatType, outendian);
			pp+=nInBlock*xnfTypeSize[outfloatType];
			nInBlock=0;
		}
		if(n==0){
			int done=0;
			fpos_t doffset=-lengths[0];
			for(int i=1;i<dim;i++){
				if(index[i]<lengths[i]-1){
					index[i]++;
					doffset+=cumulDims[i];
					done=1;break;
				}else {index[i]=0;doffset-=cumulDims[i]*(lengths[i]-1);}
			}
			if(!done ){
				if(nInBlock)TranslateData(nInBlock,buffer, infloatType, inendian,pp, outfloatType, outendian);
				break;
			}
			fseek(f,doffset*xnfTypeSize[infloatType],SEEK_CUR);
			n=lengths[0];
		}
	}
	free(buffer);
	return err;
}

int xnfReadRangeContents(xmlNodePtr cur,char *pp,NumberType floatType,int outEndian,int dim,int*dims,int * lengths,fpos_t offset){
	NumberType infloatType;int byte_order;__int64_t offsetf;
	int err=xnfDataGetInfo( cur, &infloatType,&byte_order ,&offsetf);
	if(err)return err;
	offset=offset*xnfTypeSize[infloatType]+offsetf;
	xmlChar * xnfpath= PathFromHref(cur);
	if(!xnfpath)return dataMissing;
	FILE	*f=fopen(( char * )xnfpath, "r");
	xmlFree(xnfpath);
	if(!f)return cantOpen;
	err=xnfReadRangeContentsf(f,pp, infloatType, byte_order,  floatType,outEndian, dim,dims, lengths,  offset);
	fclose(f);
	return err;
}	
static int xnfDataGetRangeContentsAlloc(xmlNodePtr cur,void**pp,NumberType floatType,int*starts,int* lengths,int* outDim, int*outdims,int doAlloc){
	int dim,dims[MaxDim];
	memset(dims,0,MaxDim*sizeof(int));
	int err=xnfGetDimensions(cur->parent,&dim, dims);
	if(err)return err;
	int size=1;
	fpos_t offset=0;
	fpos_t block=1;
	for(int i=0;i<dim;i++){
		if(starts[i]<0)return IllegalStartRange;
		if(starts[i]+lengths[i]>dims[i])return IllegalEndRange;
		size*=lengths[i];
		offset+=starts[i]*block;
		block*=dims[i];
	}
	if(!size)return 0;
	if(doAlloc)*pp=malloc(size*xnfTypeSize[floatType]);
	err=xnfReadRangeContents( cur,(char*)*pp, floatType,currentByteOrder,dim,dims, lengths,offset);
	if(err && doAlloc)free(*pp);
	int n=0;
	for(int i=0;i<dim;i++)
		if(lengths[i]>1){outdims[n]=lengths[i];n++;}
	*outDim=n;
	return err;
}

static int xnfArrayGetRangeContentsAlloc(xmlNodePtr node,int index,void**pp,NumberType floatType,int*starts,int* lengths,int* outDim, int*outdims,int doAlloc){
	xmlNodePtr data=0;
	int err=xnfGetDataNode( node,index, &data);
	if(err)return err;
	err= xnfDataGetRangeContentsAlloc(data,  pp, floatType,starts,lengths, outDim, outdims, doAlloc);
	return err;
}

int xnfArrayGetRangeContents(xmlNodePtr node,int index,int*starts,int* lengths,int* dim,int* dims,void**p,NumberType floatType){
	return xnfArrayGetRangeContentsAlloc( node,index, p, floatType,starts,lengths,dim, dims,1);
}

int xnfArrayGetRangeContentsMem(xmlNodePtr node,int index,int*starts,int* lengths,void*p,NumberType floatType){
	void* q=p;
	int dim, dims[MaxDim];
	return xnfArrayGetRangeContentsAlloc( node,index,(void**)&q,floatType,starts,lengths,&dim, dims,0);
}



